/**
 * OW2 Util
 * Copyright (C) 2008-2011 Bull S.A.S.
 * Contact: easybeans@ow2.org
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
* USA
*
* --------------------------------------------------------------------------
* $Id: OSGiDeployer.java 5993 2011-10-10 21:58:20Z cazauxj $
* --------------------------------------------------------------------------
*/

package org.ow2.util.ee.deploy.impl.osgi;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.ServiceReference;
import org.osgi.service.startlevel.StartLevel;
import org.ow2.util.archive.api.IArchiveMetadata;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.OSGiDeployable;
import org.ow2.util.ee.deploy.api.deployer.DeployerException;
import org.ow2.util.ee.deploy.impl.deployer.AbsDeployer;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * OSGi deployer that is able to deploy an OSGi Deployable
 * @author Florent BENOIT
 * @author Jeremy Cazaux
 */
public class OSGiDeployer extends AbsDeployer<OSGiDeployable> {

    /**
     * Logger.
     */
    private Log logger = LogFactory.getLog(OSGiDeployer.class);

    /**
     * Bundle Context to use for the deployment.
     */
    private BundleContext bundleContext = null;

    /**
     * List of Bundles that have been installed.
     */
    private Map<String, Bundle> installedBundles = null;

    /**
     * OSGi StartLevel service reference.
     */
    private StartLevel startLevel = null;

    /**
     * Bundle symbolic name property
      */
    public static String BUNDLE_SYMBOLIC_NAME = "Bundle-SymbolicName";

    /**
     * Bundle version property
     */
    public static String BUNDLE_VERSION = "Bundle-Version";

    /**
     * Default constructor.
     * @param bundleContext OSGi bundle context
     */
    public OSGiDeployer(final BundleContext bundleContext) {
        this.bundleContext = bundleContext;
        this.installedBundles = new HashMap<String, Bundle>();

        // Get the StartLevel service reference
        ServiceReference serviceReference = bundleContext.getServiceReference(StartLevel.class.getName());
        setStartLevel((StartLevel) bundleContext.getService(serviceReference));
    }

    /**
     * @param service the StartLevel service
     */
    public void setStartLevel(StartLevel service) {
        this.startLevel = service;
    }

    /**
     * Undeploy the given Bundle.
     * @param osgiDeployable the deployable to remove.
     * @throws DeployerException if the bundle is not undeployed.
     */
    public void doUndeploy(final IDeployable<OSGiDeployable> osgiDeployable) throws DeployerException {
        logger.debug("Undeploying {0}", osgiDeployable);

        // Bundle that will be handled
        Bundle osgiBundle = null;

        // Check if it has been installed ?
        if (!isDeployed(osgiDeployable)) {
            throw new DeployerException("Cannot undeploy the bundle '" + osgiDeployable + "' as it was not deployed.");
        }

        // get the bundle Object
        osgiBundle = installedBundles.get(getURL(osgiDeployable).toExternalForm());

        // Check found
        if (osgiBundle == null) {
            throw new DeployerException("No bundle found for the '" + osgiDeployable + "'.");
        }

        // Do not stop or uninstall if already uninstall/removed
        if (Bundle.UNINSTALLED != osgiBundle.getState()) {

            // Stop the bundle
            try {
                osgiBundle.stop();
            } catch (BundleException e) {
                throw new DeployerException("Cannot stop the bundle '" + osgiDeployable + "'.", e);
            }

            // Uninstall the bundle
            try {
                osgiBundle.uninstall();
            } catch (BundleException e) {
                throw new DeployerException("Cannot uninstall the bundle '" + osgiDeployable + "'.", e);
            }
        }

        // Remove it of installed bundle
        installedBundles.remove(getURL(osgiDeployable).toExternalForm());
    }


    /**
     * Deploy the given Bundle.
     * @param deployable the deployable to add.
     * @throws DeployerException if the Bundle is not deployed.
     */
    public void doDeploy(final IDeployable<OSGiDeployable> deployable) throws DeployerException {
        logger.debug("Deploying {0}", deployable);

        // Bundle that will be handled
        Bundle osgiBundle = null;

        // Cast the deployable only once
        OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

        // Set the initial bundle start level
        startLevel.setInitialBundleStartLevel(osgiDeployable.getStartLevel());

        // First, install the bundle
        try {
            String url = getURL(deployable).toExternalForm();
            if (osgiDeployable.isReference()) {
                url = "reference:".concat(url);
            }
            osgiBundle = bundleContext.installBundle(url);
        } catch (BundleException e) {
            // Check the bundle exception type
            if (e.getType() == BundleException.DUPLICATE_BUNDLE_ERROR) {
                logger.warn("Cannot deploy a bundle that is already installed", e.getMessage());
                return;
            } else {
                throw new DeployerException("Cannot install the bundle '" + osgiDeployable + "'.", e);
            }
        }

        if (osgiDeployable.isStart()) {
            // And then start it
            try {
                if (osgiDeployable.isStartTransient()) {
                    osgiBundle.start(Bundle.START_TRANSIENT);
                } else {
                    osgiBundle.start();
                }
            } catch (BundleException e) {
                logger.error("Cannot start the bundle ''{0}''.", deployable, e);
            }
        }

        // Add it as installed
        installedBundles.put(getURL(deployable).toExternalForm(), osgiBundle);
    }

    /**
     * Checks if the given deployable is supported by the Deployer.
     * @param deployable the deployable to be checked
     * @return true if it is supported, else false.
     */
    public boolean supports(final IDeployable<?> deployable) {
        return OSGiDeployable.class.isInstance(deployable);
    }

    /**
     * Check if the given deployable is deployable or not.
     * @param deployable the deployable to check.
     * @throws DeployerException if the deployable is not supported.
     */
    protected void check(final IDeployable<?> deployable) throws DeployerException {
        if (bundleContext == null) {
            throw new DeployerException("The deployment of the deployable'" + deployable
                    + "' is not available as there is no bundle context set on this deployer.");
        }

        super.check(deployable);
    }

    /**
     * Checks if the given deployable is deployed or not.
     * @param deployable test if a given deployable is already deployed.
     * @return true if the deployable is deployed.
     * @throws DeployerException if the undeploy operation fails.
     */
    @Override
    public boolean isDeployed(final IDeployable<OSGiDeployable> deployable) throws DeployerException {

        List<Bundle> bundles = Arrays.asList(this.bundleContext.getBundles());

        IArchiveMetadata archiveMetadata = deployable.getArchive().getMetadata();

        String symbolicName = archiveMetadata.get(BUNDLE_SYMBOLIC_NAME);
        String version = archiveMetadata.get(BUNDLE_VERSION);

        for (Bundle bundle: bundles) {

            if (bundle.getSymbolicName().equals(symbolicName) && bundle.getVersion().toString().equals(version)) {

                //if the state of the bundle is "Installed" or "Resolved", we start the bundle
                if (bundle.getState() == Bundle.INSTALLED || bundle.getState() == Bundle.RESOLVED) {
                    try {
                        bundle.start();
                    } catch (BundleException e) {
                        this.logger.error("Cannot start the bundle " + bundle.getLocation(), e);
                    }
                }

                this.installedBundles.put(getURL(deployable).toExternalForm(), bundle);

                return true;
            }

        }

        return super.isDeployed(deployable);
    }

    /**
     * Deploy the list of given Bundle.
     * @param deployables the list of deployable to add.
     * @throws DeployerException if Bundles are not deployed.
     */
    @Override
    public void doDeploy(List<IDeployable<OSGiDeployable>> deployables) throws DeployerException {

        Map<OSGiDeployable, Bundle> osgiDeployableBundleMap = new HashMap<OSGiDeployable, Bundle>();

        //first, install all bundles
        for (IDeployable deployable: deployables) {
            this.logger.debug("Deploying {0}", deployable);

            // Cast the deployable only once
            OSGiDeployable osgiDeployable = OSGiDeployable.class.cast(deployable);

            // Set the initial bundle start level
            startLevel.setInitialBundleStartLevel(osgiDeployable.getStartLevel());

            try {
                String url = getURL(osgiDeployable).toExternalForm();
                if (osgiDeployable.isReference()) {
                    url = "reference:".concat(url);
                }
                osgiDeployableBundleMap.put(osgiDeployable, this.bundleContext.installBundle(url));

            } catch (BundleException e) {
                // Check the bundle exception type
                if (e.getType() == BundleException.DUPLICATE_BUNDLE_ERROR) {
                    logger.warn("Cannot deploy a bundle that is already installed", e.getMessage());
                    return;
                } else {
                    throw new DeployerException("Cannot install the bundle '" + osgiDeployable + "'.", e);
                }
            }
        }

        //And then start all bundles
        for (Map.Entry<OSGiDeployable, Bundle> entry : osgiDeployableBundleMap.entrySet()) {

            OSGiDeployable osgiDeployable = entry.getKey();
            Bundle osgiBundle = entry.getValue();

            if (osgiDeployable.isStart()) {

                // And then start it
                try {
                    if (osgiDeployable.isStartTransient()) {
                        osgiBundle.start(Bundle.START_TRANSIENT);
                    } else {
                        osgiBundle.start();
                    }
                } catch (BundleException e) {
                    logger.error("Cannot start the bundle ''{0}''.", osgiDeployable, e);
                }
            }

            // Add it as installed
            this.installedBundles.put(getURL(osgiDeployable).toExternalForm(), osgiBundle);
        }
    }

    /**
     * Undeploy a list of Bundle without dependencies problem.
     * @param osgiDeployables The list of deployable to remove.
     * @throws DeployerException if the bundle is not undeployed.
     */
    @Override
    public void doUndeploy(final List<IDeployable<OSGiDeployable>> osgiDeployables) throws DeployerException {

        Map<OSGiDeployable, Bundle> osgiDeployableBundleMap = new HashMap<OSGiDeployable, Bundle>();

        for (IDeployable osgiDeployable: osgiDeployables) {

            this.logger.debug("Undeploying {0}", osgiDeployable);

            // Check if it has been installed ?
            if (!isDeployed(osgiDeployable)) {
                throw new DeployerException("Cannot undeploy the bundle '" + osgiDeployable + "' as it was not deployed.");
            }

            // get the bundle Object
            Bundle osgiBundle = installedBundles.get(getURL(osgiDeployable).toExternalForm());

            // Check found
            if (osgiBundle == null) {
                throw new DeployerException("No bundle found for the '" + osgiDeployable + "'.");
            }

            // Do not stop or uninstall if already uninstall/removed
            if (Bundle.UNINSTALLED != osgiBundle.getState()) {
                // Stop the bundle
                try {
                    osgiBundle.stop();
                } catch (BundleException e) {
                    throw new DeployerException("Cannot stop the bundle '" + osgiDeployable + "'.", e);
                }

                osgiDeployableBundleMap.put(OSGiDeployable.class.cast(osgiDeployable), osgiBundle);

            } else {
                // Remove it of installed bundle
                installedBundles.remove(getURL(osgiDeployable).toExternalForm());
            }
        }

        //And then uninstall all installed bundles
        for (Map.Entry<OSGiDeployable, Bundle> entry : osgiDeployableBundleMap.entrySet()) {
            OSGiDeployable osgiDeployable = entry.getKey();
            Bundle osgiBundle = entry.getValue();

            // Uninstall the bundle
            try {
                osgiBundle.uninstall();
            } catch (BundleException e) {
                throw new DeployerException("Cannot uninstall the bundle '" + osgiDeployable + "'.", e);
            }

            // Remove it of installed bundle
            installedBundles.remove(getURL(osgiDeployable).toExternalForm());
        }
    }
}
